/***************************************************************************
 *            audio.c
 *
 *  Fri Aug  6 15:17:58 2004
 *  Copyright  2004  benjie
 ****************************************************************************/

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "support.h"
#include "tocdoc.h"
#include "timeline.h"
#include "audio.h"

#ifdef WIN32
#include "audio_win32.h"
#else
#include "audio_oss.h"
#endif


// Size of one block of audio data in bytes.
#define BUF_SIZE 8192

const glong bufferLength = 0.2/*seconds*/ * 44100;

SNDFILE                 *sndFile = NULL;
gchar                   *lastFilename;
GThread                 *sndThread;
glong                   samplesWrittenSinceStart;
gboolean                hasSetStartTime;

extern tlType           *tl;

gpointer threadPlay(gpointer data);
glong playFromFile(cursorType *cursor, gchar *filename, glong start, glong framesToPlay);

gboolean startAudio(void);
glong    writeBuffer(gchar *buffer, glong bytes);
gboolean syncAudio(void);
gboolean resetAudio(void);
void     closeAudio(void);



gboolean initAudio(cursorType *cursor)
{
    cursor->playingFlag = 0;
    return TRUE;
}


glong getFileLength(gchar *filename)
{
    SNDFILE     *sndFile;
    SF_INFO     sfinfo;
    
    if (filename) {
        sndFile = sf_open_relative(filename, &sfinfo);
        if (!sndFile) {
            //perror("getFileLength");
            return teNotSound;
        }
        if (sfinfo.channels != 2 || !(sfinfo.channels & SF_FORMAT_PCM_16) ||
            sfinfo.samplerate != SAMPLES_PER_SEC) {
                sf_close(sndFile);
                return teBadSound;
        }
        sf_close(sndFile);
        return S2F(sfinfo.frames);
    }
    return teNotSound;
}


gboolean startPlaying(cursorType *cursor)
{
    if(!cursor->playingFlag) {
        
        if (!startAudio())
            return FALSE;
        
        cursor->playingFlag=1;
        cursor->stopFlag=0;
        g_timer_start(cursor->timer);
        hasSetStartTime = FALSE;
        
        sndThread = g_thread_create(&threadPlay, cursor, TRUE, NULL);
        return TRUE;
    }

    return FALSE;
}


gpointer threadPlay(gpointer data)
{
    cursorType  *cursor = data;
    tocEntry    *entry;
    glong       start, curTime, samplesToPlay, pregapSamplesToPlay, samplesPlayed;
    
    while (!cursor->stopFlag) {
        g_mutex_lock(cursor->mutex);
            cursor->changedFlag = FALSE;
        g_mutex_unlock(cursor->mutex);

        curTime = getCursorTime(cursor);
        entry = tocDocWhatIsAtTime(tl->document, S2F(curTime));
        if (entry) {
            pregapSamplesToPlay = MAX(0, F2S(entry->pregapLength_F + entry->start_F) - curTime);
            samplesToPlay = F2S(tocEntryLength(entry) + entry->start_F) - curTime - pregapSamplesToPlay;
            start = MAX(0, MAX(0, curTime - F2S(entry->start_F + entry->pregapLength_F)) + F2S(entry->offset_F));
            while (entry) {
                if (pregapSamplesToPlay > 0) {
                    samplesPlayed = playFromFile(cursor, NULL, 0, pregapSamplesToPlay);
                    if(samplesPlayed == -1) {
                        printf("playEntry - %s - playfromfile--oops\n", entry->filename);
                        closeAudio();
                        cursor->playingFlag=0;
                        return NULL;
                    }
                    if (samplesPlayed < pregapSamplesToPlay) {
                        // Probably due to (cursor.changed == TRUE)
                        //printf("shortPlay - %ld, %ld\n", samplesToPlay, samplesPlayed);
                        break;
                    }
                }
                samplesPlayed = playFromFile(cursor, entry->filename, start, samplesToPlay);
                if(samplesPlayed == -1) {
                    samplesPlayed = playFromFile(cursor, NULL, 0, samplesToPlay);
                    if (samplesPlayed == -1) {
                        printf("\nplayEntry - %s - playfromfile--oops", entry->filename);
                        closeAudio();
                        cursor->playingFlag=0;
                        return NULL;
                   }
                }
                if (samplesPlayed < samplesToPlay) {
                    // Probably due to (cursor.changed == TRUE)
                    //printf("shortPlay - %ld, %ld\n", samplesToPlay, samplesPlayed);
                    break;
                }
                if (!g_list_next(entry->item)) {
                    entry = NULL;
                    break;
                } else {
                    entry = tocEntryNext(entry);
                    samplesToPlay = F2S(entry->audioLength_F);
                    start = F2S(entry->offset_F);
                    pregapSamplesToPlay = F2S(entry->pregapLength_F);
                }
            }
        }
        if (!entry || getCursorTime(cursor) >= F2S(tocDocLength(tl->document)))
            break;
    }
    curTime = getCursorTime(cursor);
    if (curTime < F2S(tocDocLength(tl->document))) {
        samplesToPlay = F2S(tocDocLength(tl->document)) - curTime;
        hasSetStartTime = TRUE;
        if((playFromFile(cursor, NULL, 0, samplesToPlay)) == -1) {
            closeAudio();
            cursor->playingFlag=0;
            return NULL;
        }
    } else {
        ms_sleep(1000/FRAMES_PER_SEC);
    }

    syncAudio();
    curTime = getCursorTime(cursor);
    g_mutex_lock(cursor->mutex);
        cursor->startPos_S = curTime;
        cursor->playingFlag = 0;
    g_mutex_unlock(cursor->mutex);
    closeAudio();
    return NULL;
}


// returns samples played
glong playFromFile(cursorType *cursor, gchar *filename, glong start, glong samplesToPlay)
{
    gint        i;
    glong       pos = start;
    glong       lenR, lenW;
    SF_INFO     sfinfo;
    unsigned char audio_buffer[BUF_SIZE];

    //printf("%s-%ld-%ld\n", (filename)?filename:"SILENCE", start, samplesToPlay); fflush(stdout);
    
    if (filename) {
        sfinfo.format = 0;

        if(!sndFile || !lastFilename || strcmp(lastFilename, filename) != 0) {
            if(sndFile) {
                sf_close(sndFile);
            }
            sndFile = sf_open_relative(filename, &sfinfo);
            if (sndFile) {
                if (lastFilename)
                    g_free(lastFilename);
                lastFilename = g_strdup(filename);
            } else {
                if (lastFilename) {
                    g_free(lastFilename);
                    lastFilename = NULL;
                }
                perror("pff open");
            }
            if (sfinfo.channels != 2 || !(sfinfo.channels & SF_FORMAT_PCM_16) ||
                sfinfo.samplerate != SAMPLES_PER_SEC) {
                    printf("wrong file type\n");
                    if (sndFile){
                        sf_close(sndFile);
                        sndFile = NULL;
                    }
                    if (lastFilename) {
                        free(lastFilename);
                        lastFilename = NULL;
                    }
            }
        }
        pos = sf_seek(sndFile, start, SEEK_SET);
        if (pos == -1) {
            perror("seek failed");
            if (sndFile) {
                sf_close(sndFile);
                sndFile = NULL;
            }
            if (lastFilename) {
                free(lastFilename);
                lastFilename = NULL;
            }
            pos = start;
        }

    } else if (sndFile) {
        sf_close(sndFile);
        sndFile = NULL;
    }
    if(!sndFile)
        for (i=0; i<BUF_SIZE; i+=4)
            *(long *)(audio_buffer+i)=(long)0;
    
    for (;;) {
        if (cursor->stopFlag) {
            resetAudio();
            return pos-start;
        }
        if (cursor->changedFlag) {
            resetAudio();
            return pos-start;
        }
        if (pos >= start + samplesToPlay) {
            return pos-start;
        }
        
        if (sndFile) {
            lenR = sf_read_raw(sndFile, audio_buffer, MIN((samplesToPlay-(pos-start))*4, BUF_SIZE));
            if(lenR==-1) {
                perror("read error");
                sf_close(sndFile);
                return -1;
            }
            if (lenR==0) {
                return samplesToPlay;
            }
            for (; lenR < MIN((samplesToPlay-(pos-start))*4, BUF_SIZE); lenR+=4)
                *(long *)(audio_buffer+lenR)=(long)0;
        } else {
            lenR = MIN((samplesToPlay-(pos-start))*4, BUF_SIZE);
        }
        

        if (!hasSetStartTime) {
            g_mutex_lock(cursor->mutex);
                g_timer_start(cursor->timer);
            g_mutex_unlock(cursor->mutex);
            samplesWrittenSinceStart = 0;
            hasSetStartTime = TRUE;
        }

        while (samplesWrittenSinceStart >= g_timer_elapsed(cursor->timer, NULL)*SAMPLES_PER_SEC + bufferLength) {
            ms_sleep(50);
        }

        if ((lenW = writeBuffer(audio_buffer, lenR)) == -1) {
            return -1;
        }
        pos += lenW/4;
        samplesWrittenSinceStart += lenW/4;
        
        if (lenR==0)
            break;

    }
    return pos-start;
}


glong getCursorTime(cursorType *cursor)
{
    glong       startpos;

    g_mutex_lock(cursor->mutex);
    if (cursor->playingFlag){
        startpos = cursor->startPos_S;
        g_mutex_unlock(cursor->mutex);
        return startpos + (glong)(g_timer_elapsed(cursor->timer, NULL) * SAMPLES_PER_SEC);
    }
    g_mutex_unlock(cursor->mutex);

    return cursor->startPos_S;
}


void setCursorTime(cursorType *cursor, glong time)
{
    if (cursor->playingFlag) {
        cursor->startPos_S = time;
        g_timer_start(cursor->timer);
        samplesWrittenSinceStart = 0;
        cursor->changedFlag = 1;
        hasSetStartTime = FALSE;
    } else {
        cursor->startPos_S = time;
    }
}


gboolean stopPlaying(cursorType *cursor, gboolean wait)
{
    if(cursor->playingFlag) {
        cursor->stopFlag=1;
        
        if(wait) {
            g_thread_join(sndThread);
        }
        g_timer_stop(cursor->timer);
        cursor->startPos_S = F2S(cursor->lastClickFrame);
        cursor->useStaticCursor = TRUE;
        cursor->playingFlag=0;
        return TRUE;
    }
    return FALSE;
}


SNDFILE *sf_open_relative(char *filename, SF_INFO *sfinfo)
{
    gchar *basename, *dirname;
    gchar *altFilename;
    SNDFILE *SFILE;
    
    if (filename && tl && tl->document->filename && !g_file_test(filename, G_FILE_TEST_EXISTS)) {
        dirname = g_path_get_dirname(tl->document->filename);
        basename = g_path_get_basename(filename);
        altFilename = g_build_filename(dirname, basename, NULL);
        free(dirname);
        free(basename);
        if (g_file_test(altFilename, G_FILE_TEST_EXISTS)) {
            SFILE = sf_open(altFilename, SFM_READ, sfinfo);
            free(altFilename);
            return SFILE;
        }
        free(altFilename);
    }

    return sf_open(filename, SFM_READ, sfinfo);
}    


void ms_sleep(glong ms)
{
#ifdef WIN32
    _sleep(ms);
#else
    usleep(ms*1000);
#endif
}



// Sound System independent Audio Functions

gboolean startAudio(void)
{
#ifdef WIN32
    return win32_startAudio();
#else
    return oss_startAudio();
#endif
}

glong    writeBuffer(gchar *buffer, glong bytes)
{
#ifdef WIN32
    return win32_writeBuffer(buffer, bytes);
#else
    return oss_writeBuffer(buffer, bytes);
#endif
}

gboolean syncAudio(void)
{
#ifdef WIN32
    return win32_syncAudio();
#else
    return oss_syncAudio();
#endif
}

gboolean resetAudio(void)
{
#ifdef WIN32
    return win32_resetAudio();
#else
    return oss_resetAudio();
#endif
}

void closeAudio(void)
{
#ifdef WIN32
    win32_closeAudio();
#else
    oss_closeAudio();
#endif
}
