/***************************************************************************
 *            callbacks.c
 *
 *  Fri Jul 23 11:43:41 2004
 *  Copyright  2004  benjie
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <string.h>

#include "support.h"
#include "callbacks.h"
#include "tocdoc.h"
#include "timeline.h"
#include "interface.h"
#include "inspector.h"
#include "audio.h"
#include "burning.h"

static GtkFileSelection *fileselectionSave = NULL;
static GtkFileSelection *fileselectionOpen = NULL;
static GtkFileSelection *fileselectionBrowse = NULL;
static GtkFileSelection *fileselectionInsert = NULL;

gint drawCursorTag;

extern tlType       *tl;
extern GList        *tocSelection;
extern prefsStruct  prefs;


void
on_new_activate                       (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{
    tocDoc *newDoc;
    
    if (tl->document && tl->document->dirty &&
        ! AskMessage("Are you sure you want to create a new file without saving?"))
            return;
    
    newDoc = newTocDoc();
    g_mutex_lock(tl->cursor.mutex);
        if (tl->document) freeTocDoc(tl->document);
        tl->document = newDoc;
        tl->cursor.changedFlag = 1;
    g_mutex_unlock(tl->cursor.mutex);
    tocDocSetDirty(tl->document, FALSE);
    
    updateInspector();
    redrawTimeline(tl);
    tlSetScroller(tl, 0);
}


void
on_open_activate                      (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{
    if (tl->document && tl->document->dirty &&
        ! AskMessage("Are you sure you want to Close this file without saving?"))
            return;

    gtk_widget_show(GTK_WIDGET(fileselectionOpen = GTK_FILE_SELECTION(create_fileselectionOpen())));
}


void
on_save_activate                      (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{
    FILE *tocFile;
    
    if (! tl->document->dirty)
        return;

    if (tl->document->filename) {
        tocFile = fopen(tl->document->filename, "w");
        writeTocToFile(tocFile, tl->document, NULL);
        fclose(tocFile);
        tocDocSetDirty(tl->document, FALSE);
    } else {
        on_save_as_activate(menuitem, user_data);
    }
}


void
on_save_as_activate                   (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{
    gchar *name;
    
    fileselectionSave = GTK_FILE_SELECTION(create_fileselectionSave());
    name = (tl->document->filename) ? tl->document->filename : "untitled.toc";
    gtk_file_selection_set_filename(fileselectionSave, name);
    if (strcmp(name + MAX(strlen(name)-4, 0), ".toc") == 0) {
        gtk_editable_select_region(GTK_EDITABLE(fileselectionSave->selection_entry),
            0, strlen(gtk_entry_get_text(GTK_ENTRY(fileselectionSave->selection_entry)))-4);
    } else {
        gtk_editable_select_region(GTK_EDITABLE(fileselectionSave->selection_entry),
            0, strlen(gtk_entry_get_text(GTK_ENTRY(fileselectionSave->selection_entry))));
    }
    gtk_widget_show(GTK_WIDGET(fileselectionSave));
}


void
on_quit_activate                      (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{
    if (tl->document && tl->document->dirty &&
        ! AskMessage("Are you sure you want to Quit without saving?"))
            return;

    gtk_main_quit();
}


void
on_cut_activate                     (GtkMenuItem     *menuitem,
                                     gpointer         user_data)
{
    GtkWidget *focused;
    
    focused = gtk_window_get_focus(GTK_WINDOW(lookup_widget(GTK_WIDGET(menuitem), "main")));
    
    if (GTK_IS_EDITABLE(focused)) {
        gtk_editable_cut_clipboard(GTK_EDITABLE(focused));

    } else if (focused == tl->timeline) {
        g_mutex_lock(tl->cursor.mutex);
            tocDocDeleteSelection(tl->document);
            tl->cursor.changedFlag = 1;
        g_mutex_unlock(tl->cursor.mutex);
        
        updateInspector();
        redrawTimeline(tl);
        tlSetScroller(tl, -1);
    }
}


void
on_copy_activate                    (GtkMenuItem     *menuitem,
                                     gpointer         user_data)
{
    GtkWidget *focused;
    
    focused = gtk_window_get_focus(GTK_WINDOW(lookup_widget(GTK_WIDGET(menuitem), "main")));
    if (GTK_IS_EDITABLE(focused)) {
        gtk_editable_copy_clipboard(GTK_EDITABLE(focused));
    }
}


void
on_paste_activate                   (GtkMenuItem     *menuitem,
                                     gpointer         user_data)
{
    GtkWidget *focused;
    
    focused = gtk_window_get_focus(GTK_WINDOW(lookup_widget(GTK_WIDGET(menuitem), "main")));
    if (GTK_IS_EDITABLE(focused)) {
        gtk_editable_paste_clipboard(GTK_EDITABLE(focused));
    }
}


void
on_delete_activate                  (GtkMenuItem     *menuitem,
                                     gpointer         user_data)
{
    GtkWidget *focused;
    
    focused = gtk_window_get_focus(GTK_WINDOW(lookup_widget(GTK_WIDGET(menuitem), "main")));

    if (GTK_IS_EDITABLE(focused)) {
        gtk_editable_delete_selection(GTK_EDITABLE(focused));

    } else if (focused == tl->timeline) {
        if (tl->document->selection) {
            g_mutex_lock(tl->cursor.mutex);
                tocDocDeleteSelection(tl->document);
                tl->cursor.changedFlag = 1;
            g_mutex_unlock(tl->cursor.mutex);
            
            tocDocSetDirty(tl->document, TRUE);
    
            updateInspector();
            redrawTimeline(tl);
            tlSetScroller(tl, -1);
        }
    }
}


void
on_select_all_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *focused;
    GList *i;
    
    focused = gtk_window_get_focus(GTK_WINDOW(lookup_widget(GTK_WIDGET(menuitem), "main")));

    if (GTK_IS_EDITABLE(focused)) {
        gtk_editable_select_region(GTK_EDITABLE(focused), 0, -1);

    } else if (focused == tl->timeline) {
        for (i = tl->document->entries; i; i = g_list_next(i)) {
            if(!((tocEntry *)i->data)->selected)
                tocSelect(tl->document, i->data);
        }
        updateInspector();
        redrawTimeline(tl);
    }
}


void
on_about_activate                     (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{
    GtkWidget *about;
    
    about = create_aboutDialog();
    gtk_dialog_run(GTK_DIALOG(about));
    gtk_widget_destroy(about);
}


void
on_undo_activate                      (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{

}


void
on_redo_activate                      (GtkMenuItem     *menuitem,
                                       gpointer         user_data)
{

}


gboolean
on_main_delete_event                   (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    if (tl->document && tl->document->dirty &&
        ! AskMessage("Are you sure you want to Quit without saving?"))
            return TRUE;

    gtk_main_quit();
    return FALSE;
}


void
on_browse_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    gchar *filename;
    
    fileselectionBrowse = GTK_FILE_SELECTION(create_fileselectionBrowse());
    if (g_list_length(tl->document->selection) == 1) {
        filename = ((tocEntry *)((GList *)g_list_first(tl->document->selection)->data)->data)->filename;
        if (filename)
            gtk_file_selection_set_filename(fileselectionBrowse, filename);
    }
    gtk_widget_show(GTK_WIDGET(fileselectionBrowse));
}


void
on_play_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    glong curTime;
    
    tl->cursor.useStaticCursor = TRUE;
    curTime = GetActiveCursorTime(tl);
    g_mutex_lock(tl->cursor.mutex);
        setCursorTime(&(tl->cursor), curTime);
    g_mutex_unlock(tl->cursor.mutex);
    tl->cursor.useStaticCursor = FALSE;
    
    if(startPlaying(&tl->cursor)) {
        drawCursorTag = g_timeout_add(15, tlDrawCursor, tl);
    }
}


void
on_stop_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    stopPlaying(&tl->cursor, 1);
}


void
on_playStart_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    g_mutex_lock(tl->cursor.mutex);
        setCursorTime(&(tl->cursor), 0);
    g_mutex_unlock(tl->cursor.mutex);
    tl->cursor.useStaticCursor = FALSE;
    
    if(startPlaying(&tl->cursor)) {
        drawCursorTag = g_timeout_add(15, tlDrawCursor, tl);
    }
}


void
on_zoom_in_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    tlZoomIn(tl);
    redrawTimeline(tl);
}


void
on_zoom_out_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    tlZoomOut(tl);
    redrawTimeline(tl);
}


void
on_zoom_fit_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    tlZoomFit(tl);
    redrawTimeline(tl);
}


void
on_insertFile_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    gtk_widget_show(GTK_WIDGET(fileselectionInsert = GTK_FILE_SELECTION(create_fileselectionInsert())));
}


void
on_split_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    tlSplitAtCursor(tl);
}


gboolean
on_timeline_configure_event            (GtkWidget       *widget,
                                        GdkEventConfigure *event,
                                        gpointer         user_data)
{
    tlConfigure(tl);
    return TRUE;
}


gboolean
on_timeline_expose_event               (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
    tlExpose(tl, event);
	return FALSE;
}


gboolean
on_timeline_button_press_event         (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
    return tlButtonPress(tl, event);
}


gboolean
on_timeline_key_press_event            (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{
    return tlKeyPress(tl, event);
}


void
on_scroller_value_changed              (GtkRange        *range,
                                        gpointer         user_data)
{
    tlScrollerChanged(tl);
}


gboolean
on_zoomIn_button_press_event           (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
    tlZoomIn(tl);
    redrawTimeline(tl);
    return TRUE;
}


gboolean
on_zoomOut_button_press_event          (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
    tlZoomOut(tl);
    redrawTimeline(tl);
    return TRUE;
}


void
on_timeline_drag_data_received         (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{
    tlDrop(tl, x, data->data);
    gdk_drop_finish(drag_context, TRUE, time);
}


gboolean
on_timeline_destroy_event              (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    return tlDestroy(tl);
}


gboolean
on_timeline_motion_notify_event        (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
    return tlMotionNotify(tl, event);
}


gboolean
on_timeline_button_release_event       (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
    return tlButtonRelease(tl, event);
}

gboolean
on_timeline_enter_notify_event         (GtkWidget       *widget,
                                        GdkEventCrossing *event,
                                        gpointer         user_data)
{
    return tlEnterNotify(tl, event);
}


gboolean
on_timeline_leave_notify_event         (GtkWidget       *widget,
                                        GdkEventCrossing *event,
                                        gpointer         user_data)
{
    return tlLeaveNotify(tl, event);
}


void
on_insert_track_break_activate         (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    glong newBreakPos;
    
    if (tl->breakSelected != -1 && tl->cursor.playingFlag) tl->cursor.useStaticCursor = FALSE;
    newBreakPos = S2F(GetActiveCursorTime(tl));
    tocDocAddBreak(tl->document, newBreakPos);
    tocClearSelection(tl->document);
    tl->breakSelected = newBreakPos;
    tocDocSetDirty(tl->document, TRUE);
    redrawTimeline(tl);
}


void
on_burn_cd_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    FILE        *tocFile;
    gchar       filenameTemplate[] = "daoism.tmp-XXXXXX";
    gchar       *filename;

    tocFile = fdopen(g_file_open_tmp(filenameTemplate, &filename, NULL), "w");

    if (!tocFile) return;

#ifdef WIN32
    writeTocToFile(tocFile, tl->document, &cygpathize);
#else
    writeTocToFile(tocFile, tl->document, NULL);
#endif
    fclose(tocFile);
    
    startBurn(filename);
}


void
on_auto_pregap_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    prefs.autoPregap = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(menuitem));
    savePrefs();
}


void
on_insert_silence_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    tlInsertSilence(tl);
}


void
on_fileselectionOpen_response          (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
    tocDoc *newDoc;
	gchar *filename;
    
    if (response_id == -5) {
        filename = (gchar *)gtk_file_selection_get_filename((GtkFileSelection*)fileselectionOpen);
        newDoc = openTocDoc(filename);
        if (!newDoc) newDoc = newTocDoc();
        g_mutex_lock(tl->cursor.mutex);
            if(tl->document) freeTocDoc(tl->document);
            tl->document = newDoc;
            tl->cursor.changedFlag = 1;
            tocDocSetDirty(tl->document, FALSE);
        g_mutex_unlock(tl->cursor.mutex);
        tlZoomFit(tl);
    }
    gtk_widget_hide(GTK_WIDGET(fileselectionOpen));
    gtk_widget_destroy(GTK_WIDGET(fileselectionOpen));
    
    updateInspector();
    redrawTimeline(tl);
    tlSetScroller(tl, 0);
}


void
on_fileselectionSave_response          (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
	gchar *filename;
    FILE  *tocFile;
    
    if (response_id == -5) {
        filename = (gchar *)gtk_file_selection_get_filename((GtkFileSelection*)fileselectionSave);
        if (!g_file_test(filename, G_FILE_TEST_EXISTS) ||
            (g_file_test(filename, G_FILE_TEST_IS_REGULAR) &&
            AskMessage("A file with that name already exists.  Replace it?"))) {
            tocFile = fopen(filename, "w");
            writeTocToFile(tocFile, tl->document, NULL);
            fclose(tocFile);
            tl->document->filename = g_strdup(filename);
            tocDocSetDirty(tl->document, FALSE);
        } else {
            // leave Save box open
            return; 
        }
    }
    gtk_widget_hide(GTK_WIDGET(fileselectionSave));
    gtk_widget_destroy(GTK_WIDGET(fileselectionSave));
}


void
on_fileselectionBrowse_response        (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
	gchar   *filename;
    GList   *item;
    glong   length;
    
    if (response_id == -5) {
        filename = (gchar *)gtk_file_selection_get_filename((GtkFileSelection*)fileselectionBrowse);
        length = getFileLength(filename);        
        if (length > 0) {
            item = g_list_first(tl->document->selection);
            g_mutex_lock(tl->cursor.mutex);
            while (item) {
                tocEntrySetFilename((tocEntry *)((GList *)item->data)->data, filename);
                item = g_list_next(item);
            }
            g_mutex_unlock(tl->cursor.mutex);
            tocDocSetDirty(tl->document, TRUE);
        }
    }
    gtk_widget_hide(GTK_WIDGET(fileselectionBrowse));
    gtk_widget_destroy(GTK_WIDGET(fileselectionBrowse));

    updateInspector();
    redrawTimeline(tl);
}


void
on_fileselectionInsert_response        (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
    gchar   *filename;
    glong   length;
    
    if (response_id == -5) {
        filename = (gchar *)gtk_file_selection_get_filename((GtkFileSelection*)fileselectionInsert);
        length = getFileLength(filename);        
        if (length > 0) {
            tlInsertFile(tl, filename);
            tocDocSetDirty(tl->document, TRUE);
        }
    }
    gtk_widget_hide(GTK_WIDGET(fileselectionInsert));
    gtk_widget_destroy(GTK_WIDGET(fileselectionInsert));
    
    updateInspector();
    redrawTimeline(tl);
    tlSetScroller(tl, 0);
}


void
on_properties_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget   *prefDialog;
    gchar       lengthStr[8];
    
    prefDialog = create_prefDialog();
    
    // Setup Values
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget(
        prefDialog, "pregapCheckbutton")), prefs.autoPregap);
    snprintf(lengthStr, 7, "%d", prefs.maxCdLength);
    gtk_entry_set_text(GTK_ENTRY(lookup_widget(prefDialog, "maxLengthEntry")), lengthStr);
    gtk_entry_set_text(GTK_ENTRY(lookup_widget(prefDialog, "commandEntry")), prefs.cdrdaoCmdStr);
    gtk_entry_set_text(GTK_ENTRY(lookup_widget(prefDialog, "deviceEntry")), prefs.deviceStr);
    gtk_entry_set_text(GTK_ENTRY(lookup_widget(prefDialog, "driverEntry")), prefs.driverStr);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget(
        prefDialog, "ejectCheckbutton")), prefs.eject);
    
    if (gtk_dialog_run(GTK_DIALOG(prefDialog)) == -5) {
        // Apply Values
        g_free(prefs.deviceStr);
        g_free(prefs.driverStr);
        prefs.autoPregap = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(
            lookup_widget(prefDialog, "pregapCheckbutton")));
        sscanf(gtk_entry_get_text(GTK_ENTRY(
            lookup_widget(prefDialog, "maxLengthEntry"))), "%d", &prefs.maxCdLength);
        prefs.cdrdaoCmdStr = g_strdup(gtk_entry_get_text(GTK_ENTRY(
            lookup_widget(prefDialog, "commandEntry"))));
        prefs.deviceStr = g_strdup(gtk_entry_get_text(GTK_ENTRY(
            lookup_widget(prefDialog, "deviceEntry"))));
        prefs.driverStr = g_strdup(gtk_entry_get_text(GTK_ENTRY(
            lookup_widget(prefDialog, "driverEntry"))));
        prefs.eject = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(
            lookup_widget(prefDialog, "ejectCheckbutton")));
        savePrefs();

        gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(lookup_widget(
            GTK_WIDGET(menuitem), "auto_pregap")), prefs.autoPregap);
    }
    gtk_widget_destroy(prefDialog);
}

void
on_timeline_realize                    (GtkWidget       *widget,
                                        gpointer         user_data)
{

}
