/***************************************************************************
 *            tocdoc.c
 *
 *  Fri Jul 23 11:43:41 2004
 *  Copyright  2004  benjie
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "tocdoc.h"
#include "inspector.h"
#include "timeline.h"
#include "thumbnail.h"
#include "audio.h"

extern tlType *tl;
extern prefsStruct prefs;

#ifdef WIN32
#include <windows.h>
extern gchar *daoismDir;
#endif


gint compareTocEntries(tocEntry *a, tocEntry *b) {return a->start_F - b->start_F;}
gint compareSelEntries(GList *a, GList *b) {return ((tocEntry *)a->data)->start_F - ((tocEntry *)b->data)->start_F;}
gint compareTrackBreaks(gconstpointer a, gconstpointer b) {return (glong)a - (glong)b;}

void writeTocEntryToFile(FILE *tocFile, gchar *filename, glong offset,
                         glong pregap, glong length, gboolean newTrack,
			 renamingFunc renamer);


tocDoc *newTocDoc(void)
{
	tocDoc *newTocDoc;
	
	newTocDoc = (tocDoc *)malloc(sizeof(tocDoc));
    newTocDoc->entries = NULL;
	newTocDoc->selection = NULL;
    newTocDoc->filename = NULL;
    newTocDoc->dirty = FALSE;

	return newTocDoc;
}


tocDoc *openTocDoc(gchar *file)
{
	tocDoc *doc;

	doc = newTocDoc();
    doc->filename = g_strdup(file);
	if (!readTocFromFile(doc, file)) {
        freeTocDoc(doc);
        return NULL;
    }
    return doc;
}


gboolean readTocFromFile(tocDoc *doc, gchar *file)
{
	FILE        *tocfile;
    tocEntry    *entry=NULL;
	gint        n;
	gchar       buf[1024];
    gchar       args[4][1024];
    glong       offset, length;
	gboolean    addBreak;
    glong       pregap;
    
    addBreak = TRUE;
    pregap = 0;
	tocfile = fopen(file, "r");
    if (!tocfile) return FALSE;
    
    // Check for CD_DA header
	fgets(buf, 1024, tocfile);
    n = sscanf(buf, "%s", args[0]);
    if (n != 1 || strcasecmp(args[0], "CD_DA") != 0) {
        printf("Not a valid toc file. " "(doesn't start with 'CD_DA')\n");
        fclose(tocfile);
        return FALSE;
    }
    
	while (fgets(buf, 1024, tocfile)){
		n = sscanf(buf, "%s", args[0]);
		if (n==1 && strcasecmp(args[0], "FILE")==0){
    		n = sscanf(buf, "%s \"%[^\"]\" %s %s\n", args[0], args[1], args[2], args[3]);
            offset = cdTimeToGint(args[2]);
            if (n==3 || n==4) {
                offset = cdTimeToGint(args[2]);
                length = (n==3) ? -1 : cdTimeToGint(args[3]);
                if (entry && entry->filename && strcmp(entry->filename,args[1])==0 && 
                    offset == entry->offset_F + entry->audioLength_F && pregap == 0) {
                        tocDocResizeEntry(doc, entry, (length == -1) ? -1 : (entry->audioLength_F + length));
                        if (addBreak)
                            entry->breaks = g_list_append(entry->breaks, (gpointer)offset - entry->offset_F);
                        addBreak = FALSE;
                        pregap = 0;
                } else {
                    entry = tocDocNewEntry(doc, args[1], offset, (addBreak) ? pregap : 0, length, addBreak);
                    tocDocAppendEntry(doc, entry);
                    addBreak = FALSE;
                    pregap = 0;
                }
            }
        } else if (n==1 && strcasecmp(args[0], "SILENCE")==0) {
    		n = sscanf(buf, "%s %s\n", args[0], args[1]);
            if (n==2) {
                entry = tocDocNewEntry(doc, NULL, 0, 0, cdTimeToGint(args[1]), addBreak);
                tocDocAppendEntry(doc, entry);
                addBreak = FALSE;
            }
        } else if (n==1 && strcasecmp(args[0], "PREGAP")==0) {
    		n = sscanf(buf, "%s %s\n", args[0], args[1]);
            if (n==2) {
                pregap = cdTimeToGint(args[1]);
            }
        } else if (n==1 && strcasecmp(args[0], "TRACK")==0) {
    		n = sscanf(buf, "%s %s\n", args[0], args[1]);
            if (n==2 && strcasecmp(args[1], "AUDIO")==0) {
                addBreak = TRUE;
                pregap = 0;
            }
        } else if (n>0 && args[0][0] != '#') {
            printf("Not a valid toc file. " "Couldn't understand line: %s", buf);
            fclose(tocfile);
            return FALSE;
        }
	}
	fclose(tocfile);
    return TRUE;
}


tocDoc *cloneTocDoc(tocDoc *doc)
{
    tocDoc *newDoc;
    GList  *item;
    GList  *breakItem;
    tocEntry *entry;

    newDoc = newTocDoc();

    item = doc->entries;
    while (item) {
        entry = (tocEntry *)item->data;

	tocDocAppendEntry(newDoc, tocDocNewEntry(newDoc, entry->filename, 
						 entry->offset_F,
						 entry->pregapLength_F,
						 entry->audioLength_F, FALSE));
	breakItem = entry->breaks;
	entry = (tocEntry *)(g_list_last(newDoc->entries)->data);
	while (breakItem) {
	    entry->breaks = g_list_append(entry->breaks, breakItem->data);
	    breakItem = g_list_next(breakItem);
	}
	item = g_list_next(item);
    }

    return newDoc;
}


void freeTocDoc(tocDoc *doc)
{
    tocEntry    *entry;

    tocClearSelection(doc);
	while (doc->entries) {
        entry = g_list_first(doc->entries)->data;
		if (entry->filename)
            g_free(entry->filename);
        
        if (entry->thumbnail)
            freeThumbnail(entry);
        
    	while ((entry->breaks = g_list_remove_link(entry->breaks,
                g_list_first(entry->breaks))) != NULL);
        g_free(entry);
        doc->entries = g_list_remove_link(doc->entries, g_list_first(doc->entries));
    }

    g_free(doc);
}


glong tocDocLength(tocDoc *doc)
{
    tocEntry *last;
    
    if(doc) {
        if(doc->entries) {
            last = g_list_last(doc->entries)->data;
            return last->start_F + tocEntryLength(last);
        } else {
            return 0;
        }
    } else {
        return 0;
    }
}


glong tocEntryLength(tocEntry *entry)
{
    return entry->audioLength_F + entry->pregapLength_F;
}


void writeTocToFile(FILE *tocFile, tocDoc *doc, renamingFunc renamer)
{
    tocEntry    *entry;
    GList       *breakItem;
    glong       lastBreakPos, breakPos;
    gboolean    newTrack;
    
    if (!doc->entries || !tocFile) return;

    fprintf(tocFile, "CD_DA\n");
    entry = doc->entries->data;
    
    while (entry) {
        lastBreakPos = 0;
        breakItem = entry->breaks;
        while (1) {
            breakPos = (breakItem) ? (glong)breakItem->data : entry->audioLength_F;
            
            if (breakPos != lastBreakPos) {
                newTrack = ((entry->start_F + lastBreakPos == 0) || (entry->breaks && ((glong)entry->breaks->data) == 0 && lastBreakPos == 0) || (lastBreakPos > 0));
                writeTocEntryToFile(tocFile, entry->filename,
                    lastBreakPos + entry->offset_F, (lastBreakPos==0) ? entry->pregapLength_F : 0,
                    breakPos - lastBreakPos, newTrack, renamer);
            }
            
            if (!breakItem) break;
            breakItem = g_list_next(breakItem);
            lastBreakPos = breakPos;
        }
        entry = tocEntryNext(entry);
    }
}


void writeTocEntryToFile(FILE *tocFile, gchar *filename, glong offset,
                         glong pregap, glong length, gboolean newTrack,
			 renamingFunc renamer)
{
    gchar       offsetStr[32], pregapStr[32], lengthStr[32];
    gchar *renamed;

    gintToCdTime(offsetStr, offset);
    gintToCdTime(lengthStr, length);

    if (newTrack) {
        fprintf(tocFile, "\nTRACK AUDIO\n");
    }
    if (pregap > 0) {
        gintToCdTime(pregapStr, pregap);
        fprintf(tocFile, "PREGAP %s\n", pregapStr);            
    }

    if (filename) {
      if (renamer) {
	renamed = renamer(filename);
        fprintf(tocFile, "FILE \"%s\" %s %s\n",
		renamed, offsetStr, lengthStr);
	g_free(renamed);
      } else {
        fprintf(tocFile, "FILE \"%s\" %s %s\n",
		filename, offsetStr, lengthStr);
      }
    } else {
        fprintf(tocFile, "SILENCE %s\n", lengthStr);
    }
}


tocEntry *tocDocNewEntry(tocDoc *doc, gchar *filename, glong offset, glong pregap, glong length, gboolean addBreak)
{
    tocEntry *newEntry;

    newEntry = g_malloc(sizeof(tocEntry));
    newEntry->filename = NULL;
    newEntry->thumbnail = NULL;
    tocEntrySetFilename(newEntry, filename);
    newEntry->start_F = -1;
    newEntry->offset_F = MAX(0, offset);
    newEntry->item = NULL;
    newEntry->pregapLength_F=0;
    tocDocResizeEntry(doc, newEntry, length);
    tocDocResizePregap(doc, newEntry, pregap);
    newEntry->selected = FALSE;
    if (addBreak)
        newEntry->breaks = g_list_append(NULL, (glong)0);
    else
        newEntry->breaks = NULL;
    
    return newEntry;
}


void tocEntrySetFilename(tocEntry *entry, gchar *filename)
{
    if (entry) {
        if (entry->filename && filename && strcmp(filename, entry->filename)==0)
            return;
        
        if (entry->filename) {
            g_free(entry->filename);
            entry->filename = NULL;
        }
        
        if (entry->thumbnail)
            freeThumbnail(entry);
        
        if (filename) {
            if (!entry->filename) {
                entry->filename = g_strdup(filename);
            }
            entry->fileLength_F = getFileLength(entry->filename);
            if (entry->fileLength_F == teNotSound) entry->fileType = teNotSound;
            else if (entry->fileLength_F == teBadSound) entry->fileType = teBadSound;
            else if (entry->fileLength_F < 0) entry->fileLength_F = 0;
            else {
                entry->fileType = teAudio;
                createThumbnail(entry);
            }
        } else {
            entry->filename = NULL;
            entry->fileLength_F = 0;
            entry->fileType = teSilence;
        }
    }
}


tocEntry *tocDocInsertEntryBefore(tocDoc *doc, tocEntry *newEntry, tocEntry *entry)
{
    GList *item;
    
    if (!entry)
        return tocDocAppendEntry(doc, newEntry);
    
    newEntry->start_F = entry->start_F;
    item = g_list_find(doc->entries, entry);
    if (!item){
        return NULL;
    }
    doc->entries = g_list_insert_before(doc->entries, item, newEntry);
    ((tocEntry *)g_list_previous(item)->data)->item = g_list_previous(item);
	while (item) {
        ((tocEntry *)item->data)->start_F += tocEntryLength(newEntry);
        item = g_list_next(item);
    }
    
    return newEntry;
}


void tocDocMoveSelectionBefore(tocDoc *doc, tocEntry *toEntry)
{
    GList       *newSel = NULL;
    GList       *selItem;
    tocEntry    *moveEntry;
    
    selItem = g_list_find(doc->entries, toEntry);
    while (selItem) {
        toEntry = selItem->data;
        if (!toEntry->selected) break;
        selItem = g_list_next(selItem);
    }
    selItem = doc->selection;
    while (selItem) {
        moveEntry = ((GList *)selItem->data)->data;
        selItem = g_list_next(selItem);
        
        if (moveEntry != toEntry) {
            tocDocRemoveEntry(doc, moveEntry);
            if (toEntry) {
                tocDocInsertEntryBefore(doc, moveEntry, toEntry);
            } else {
                tocDocAppendEntry(doc, moveEntry);
            }
            newSel = g_list_append(newSel, moveEntry);
        }
    }
    while (newSel) {
        moveEntry = newSel->data;
        tocSelect(doc, moveEntry);
        newSel = g_list_remove(newSel, moveEntry);
    }
}


tocEntry *tocDocAppendEntry(tocDoc *doc, tocEntry *newEntry)
{
    if (doc->entries)
        newEntry->start_F = ((tocEntry *)g_list_last(doc->entries)->data)->start_F +
            ((tocEntry *)g_list_last(doc->entries)->data)->audioLength_F +
            ((tocEntry *)g_list_last(doc->entries)->data)->pregapLength_F;
    else
        newEntry->start_F = 0;
    doc->entries = g_list_append(doc->entries, newEntry);
    ((tocEntry *)g_list_last(doc->entries)->data)->item = g_list_last(doc->entries);
   
    return newEntry;
}


void tocDocResizeEntry(tocDoc *doc, tocEntry *entry, glong length)
{
    GList *item, *delItem;
    glong pos;
    
    if (!entry) return;
    
    if (length > -1) {
        entry->audioLength_F = MAX(1, length);
    } else {
        if (entry->fileLength_F == -1)
            entry->audioLength_F = 750;
        else
            entry->audioLength_F = entry->fileLength_F;
    }

    pos = entry->start_F + tocEntryLength(entry);
    item = g_list_find(doc->entries, entry);
    if (!item) return;

    while ((item = g_list_next(item)) != NULL) {
        ((tocEntry *)item->data)->start_F = pos;
        pos += tocEntryLength((tocEntry *)item->data);
    }
    
    item = entry->breaks;
    while (item) {
        if ((glong)item->data >= entry->audioLength_F) {
            tocDocAddBreak(doc, (glong)item->data + entry->start_F + entry->pregapLength_F);
            delItem = item;
            item = g_list_next(item);
            entry->breaks = g_list_remove_link(entry->breaks, delItem);
        } else {
            item = g_list_next(item);
        }
    }
}


void tocDocResizePregap(tocDoc *doc, tocEntry *entry, glong pregap)
{
    if (!entry) return;
    
    entry->pregapLength_F = MAX(0, pregap);
    tocDocResizeEntry(doc, entry, entry->audioLength_F);
}


void tocDocSplitEntry(tocDoc *doc, tocEntry *entry, glong split)
{
    tocEntry    *newEntry;
    glong       localSplit;
    
    localSplit = split - entry->start_F;

    if (tocEntryLength(entry) == localSplit || localSplit == 0)
        return;
    
    if (localSplit < entry->pregapLength_F) {
        // Split the pregap
        newEntry = tocDocNewEntry(doc, NULL, 0, 0, localSplit, FALSE);
        tocDocInsertEntryBefore(doc, newEntry, entry);
        tocDocResizePregap(doc, entry, entry->pregapLength_F - localSplit);
    } else {
        // Split the actual track
        localSplit -= entry->pregapLength_F;
        newEntry = tocDocNewEntry(doc, entry->filename, entry->offset_F + localSplit, 0,
                                  entry->audioLength_F - localSplit, FALSE);
        tocDocInsertEntryBefore(doc, newEntry, tocEntryNext(entry));
        tocDocResizeEntry(doc, entry, localSplit);
    }
}


void tocDocRemoveEntry(tocDoc *doc, tocEntry *entry)
{
    GList *item;
    
    tocDeselect(doc, entry);

    item = g_list_find(doc->entries, entry);
    if (!item){
        return;
    }
    while ((item = g_list_next(item)) != NULL) {
        ((tocEntry *)item->data)->start_F -= tocEntryLength(entry);
    }
    doc->entries = g_list_remove(doc->entries, entry);
}


void tocDocDeleteEntry(tocDoc *doc, tocEntry *entry)
{
    tocDocRemoveEntry(doc, entry);
    if (entry->thumbnail)
        freeThumbnail(entry);
    if (entry->filename)
        g_free(entry->filename);
    g_free(entry);
}


gint tocEntryCountBreaks(tocEntry *entry)
{
    return g_list_length(entry->breaks);
}


void tocDocAddBreak(tocDoc *doc, glong time)
{
    tocEntry    *entry;
    
    entry = tocDocWhatIsAtTime(doc, time);
    if (entry && !g_list_find(entry->breaks, (gpointer)(time - entry->start_F - entry->pregapLength_F)) &&
        time - entry->start_F - entry->pregapLength_F >= 0)
        entry->breaks = g_list_insert_sorted(entry->breaks, 
            (gpointer)(time - entry->start_F - entry->pregapLength_F), compareTrackBreaks);
}


void tocDocRemoveBreak(tocDoc *doc, glong time)
{
    tocEntry    *entry;
    
    entry = tocDocWhatIsAtTime(doc, time);
    if (entry)
        entry->breaks = g_list_remove(entry->breaks, (gpointer)(time - entry->start_F - entry->pregapLength_F));
}


glong tocDocFindClosestBreak(tocDoc *doc, glong time)
{
    tocEntry    *entry;
    GList       *item, *breakItem;
    glong       nextTime;
    glong       previousTime;

    entry = tocDocWhatIsAtTime(doc, time);
    if (!entry)
        return -1;

    previousTime = -1;
    item = entry->item;
    while (item && previousTime == -1) {
        breakItem = g_list_last(((tocEntry *)item->data)->breaks);
        while (breakItem && previousTime== -1) {
            if (((tocEntry *)item->data)->start_F + 
                ((tocEntry *)item->data)->pregapLength_F + (glong)breakItem->data <= time) {
                previousTime = ((tocEntry *)item->data)->start_F + 
                    ((tocEntry *)item->data)->pregapLength_F + (glong)breakItem->data;
            }
            breakItem = g_list_previous(breakItem);
        }
        item = g_list_previous(item);
    }
    
    nextTime = -1;
    item = entry->item;
    while (item && nextTime == -1) {
        breakItem = g_list_first(((tocEntry *)item->data)->breaks);
        while (breakItem && nextTime== -1) {
            if (((tocEntry *)item->data)->start_F + 
                ((tocEntry *)item->data)->pregapLength_F + (glong)breakItem->data >= time) {
                nextTime = ((tocEntry *)item->data)->start_F + 
                    ((tocEntry *)item->data)->pregapLength_F + (glong)breakItem->data;
            }
            breakItem = g_list_next(breakItem);
        }
        item = g_list_next(item);
    }

    if (nextTime == -1 && previousTime == -1)
        return -1;
    else if (nextTime == -1)
        return previousTime;
    else if (previousTime == -1)
        return nextTime;
    else if (nextTime - time < time - previousTime)
        return nextTime;
    else
        return previousTime;
}


tocEntry *tocDocWhatIsAtTime(tocDoc *doc, glong t)
{
    GList *item;
    tocEntry *entry;
    
    item = doc->entries;
	while (item) {
        entry = (tocEntry *)item->data;
        if (entry->start_F <= t  &&
            entry->start_F + tocEntryLength(entry) > t)
                return entry;
        if (entry->start_F > t)
            return NULL;
        item = g_list_next(item);
    }
    return NULL;
}


void tocDocDeleteSelection(tocDoc *doc)
{
	while (doc->selection) {
		tocDocDeleteEntry(doc, (tocEntry *)((GList *)g_list_first(doc->selection)->data)->data);
    }
}


void tocClearSelection(tocDoc *doc)
{
	while (doc->selection) {
        ((tocEntry *)((GList *)g_list_first(doc->selection)->data)->data)->selected = FALSE;
        doc->selection = g_list_remove(doc->selection, g_list_first(doc->selection)->data);
    }
}


// Add entry to selection
void tocSelect(tocDoc *doc, tocEntry *entry)
{
    GList *item;

    item = g_list_find(doc->entries, entry);

    if (item && !entry->selected) {
        doc->selection = g_list_insert_sorted(doc->selection, item, 
            (GCompareFunc)compareSelEntries);
        entry->selected = TRUE;
    }
}


// Remove entry from selection
void tocDeselect(tocDoc *doc, tocEntry *entry)
{
    GList *item;

    item = g_list_find(doc->entries, entry);

    doc->selection = g_list_remove(doc->selection, item);
    entry->selected = FALSE;
}


tocEntry *tocEntryNext(tocEntry *entry)
{
    if (entry && entry->item && g_list_next(entry->item))
        return g_list_next(entry->item)->data;
    
    return NULL;
}


tocEntry *tocEntryPrevious(tocEntry *entry)
{
    if (entry && entry->item && g_list_previous(entry->item))
        return g_list_previous(entry->item)->data;
    
    return NULL;
}


void tocDocSetDirty(tocDoc *doc, gboolean dirty)
{
    doc->dirty = dirty;
    if (dirty) {
        gtk_widget_set_sensitive(lookup_widget(tl->timeline,"save"), TRUE);
        gtk_widget_set_sensitive(lookup_widget(tl->timeline,"save2"), TRUE);
    } else {
        gtk_widget_set_sensitive(lookup_widget(tl->timeline,"save"), FALSE);
        gtk_widget_set_sensitive(lookup_widget(tl->timeline,"save2"), FALSE);
    }
}


void openPrefs(void)
{
    FILE        *prefFile;
    gchar       line[1024];
    gchar       arg[1024];
    gint        num;
    const gchar *home;
    
#ifdef WIN32
    snprintf(line, sizeof(line), "%s\\daoismrc", daoismDir);
#else
    if ((home = getenv("HOME")) == NULL) {
        fprintf(stderr, "Home directory location unknown\n");
        return;
    }    
    snprintf(line, sizeof(line), "%s/.daoism", home);
#endif

    prefFile = fopen(line, "r");    

    prefs.autoPregap = TRUE;
    prefs.maxCdLength = 80;
    prefs.cdrdaoCmdStr = g_strdup("/usr/bin/cdrdao");
    prefs.deviceStr = g_strdup("/dev/cdrecorder");
    prefs.driverStr = g_strdup("generic-mmc");
    prefs.eject = TRUE;
    
    if (!prefFile) {
        savePrefs();
        return;
    }

    while (fgets(line, sizeof(line), prefFile)) {
        if (sscanf(line, "autoPregap %s", arg) == 1) {
            prefs.autoPregap = (g_strcasecmp(arg, "Yes") == 0);
        } else if (sscanf(line, "maxCdLength %d", &num) == 1) {
            prefs.maxCdLength = num;
        } else if (sscanf(line, "cdrdaoCommand %s", arg) == 1) {
            free(prefs.cdrdaoCmdStr);
            prefs.cdrdaoCmdStr = g_strdup(arg);
        } else if (sscanf(line, "cdrdaoDevice %s", arg) == 1) {
            free(prefs.deviceStr);
            prefs.deviceStr = g_strdup(arg);
        } else if (sscanf(line, "cdrdaoDriver %s", arg) == 1) {
            free(prefs.driverStr);
            prefs.driverStr = g_strdup(arg);
        } else if (sscanf(line, "cdrdaoEject %s", arg) == 1) {
            prefs.eject = (g_strcasecmp(arg, "Yes") == 0);
        }
    }
    fclose(prefFile);
}


void savePrefs(void)
{
    FILE        *prefFile;
    gchar       line[1024];
    const char  *home;
    
#ifdef WIN32
    snprintf(line, sizeof(line), "%s\\daoismrc", daoismDir);
#else
    if ((home = getenv("HOME")) == NULL) {
        fprintf(stderr, "Home directory location unknown\n");
        return;
    }
    snprintf(line, sizeof(line), "%s/.daoism", home);
#endif

    prefFile = fopen(line, "w");
    if (!prefFile) {
        printf("Couldn't open prefs file\n");
        return;
    }
    fprintf(prefFile, "autoPregap %s\n", (prefs.autoPregap) ? "Yes" : "No");
    fprintf(prefFile, "maxCdLength %d\n", prefs.maxCdLength);
    fprintf(prefFile, "cdrdaoCommand %s\n", prefs.cdrdaoCmdStr);
    fprintf(prefFile, "cdrdaoDevice %s\n", prefs.deviceStr);
    fprintf(prefFile, "cdrdaoDriver %s\n", prefs.driverStr);
    fprintf(prefFile, "cdrdaoEject %s\n", (prefs.eject) ? "Yes" : "No");

    fclose(prefFile);
}
